/** Spex - A 'Simple Promela EXplorer' for TorX
  * Built by:	J. v. Yperen (JvY)
  * Version:	0.7.9b 
  * Date: 	15-03-07
  *
  * For tool documentation see:  
  *             http://www.purl.org/net/torx/                                      
  * For contact with the maintainers: torx@cs.utwente.nl           
  */

/* Global TODO elements:
   - Switch to hashmap for storing states, where the State (incl state vector) is key 
     (quicker to find duplicate(s))
   - Always redirect stdout for pan
*/

/* Known issues:
   - If two states are identical, except for the (remembered) value of an uninstantiated state,
     they will not be considered identical.
   - No state will be considered identical to the initial state.
   - States that have the same hash value, but different Promela labels are considered equal, and carry the first
     label encountered. This is by design (to reduce storage space), but can be overriden by defining "PROM_LABELS_MUST_MATCH".
     Different spex-labels always make states considered to be inequal.
*/

#include <stdio.h>
#include <stdlib.h>
#ifndef SPEX
#  define SPEX
#endif
// Use BFS algorithm of PAN
#define BFS
// Don't compress states in PAN (we might need them to compare with)
#define NOCOMP

#include "spex_mtab.h"

#include "spexIO.h"

// Include pan.c (Yes, I know this isn't clean. But we need access to the
//                BFS structs, which happen to be defined in the .c file.)
#include "pan.c"

#include "spex.h"

#include "incr_string.h"


// Do we use the Domain functions during instantiation?
#ifdef DOMAIN_IO
#include "spexDomain.h"
// Rest of the includes as well?
#endif


/* Define structures */
typedef struct indexStore {
	int index;
	struct indexStore *next;
} index_Store;

typedef struct trailStore {
	char flags;
	index_Store *first_index;
	index_Store *last_index;
//	char *trans_text;
//	Var_Store *trans_vars;
	BFS_Trail *trail;
	void *hash_loc;
} Trail_Store;
/* Flags for the trail storage
   flags&1 = dirty bit (this Trail has been expanded before)
   //flags&2 = uninstantiated bit (this Trail has to be instantiated and cannot be expanded)
   //flags&4 = ...
*/

/* What makes a state unique?

   The state vector (containing all variables) largely indicates the uniqueness of a state.
   Each found state vector will have its own trailStore to keep it in.
   However, we also have to consider free variables, and different instantiated output states.
   Output states may have different labels, but they do have the same state vector.
   (Ex. output "x + 1" and "x - 1". Different labels, same underlying state (vector).)
   
   This brings me to putting labels and free variables into an "Entry" structure.
*/

typedef struct entry {
//	char flags;
	char *trans_text;
	Var_Store *trans_vars;
	Trail_Store *store;
} Entry;

/* Define functions */
index_Store *new_index_store();
void free_index_store(index_Store *);
index_Store *push_index(index_Store **, int);
void free_indices(index_Store **);
void delete_state(int state_index);
void delete_trail(Trail_Store*, int);
void free_trail(Trail_Store*);

int interact();
int handle_line(char *);
int explorer_interface();

int cmd_unknown(const char *);
int cmd_quit(const char *);
int cmd_reset(const char *);
int cmd_explore(const char *);
int cmd_instantiate(const char *);
int cmd_delete(char *);

void explore_callback(void);
void delete_state(int);



/* Define constants and global variables */
// Number of states we can store
//#ifndef STORAGE_SIZE
//#define STORAGE_SIZE 2048
//#endif
//// Longest line we can read from the stdin
//#ifndef MAX_LINE_LENGTH
//#define MAX_LINE_LENGTH 1024
//#endif

static char explore_cb = 0; // Flag for callback after an "explore" command
static index_Store *explore_cb_indices = NULL; // Indices to be printed at the callback after an "explore" command
static char explore_known = 0; // Flag for exploring a previously explored state

#ifdef PREFER_END_TRANS
static int explore_end_trans = 0; // Index for an -end- transition, which should be prioritized.
#endif

// Define at which interval of Spex states information about the Spin component should be printed.
#ifdef SPIN_STATS
#ifndef SPIN_MOD_STATES
#define SPIN_MOD_STATES 20
#endif
#endif

//static Trail_Store *storage[STORAGE_SIZE]; // Global storage for (unique) states
//static Entry *storage[STORAGE_SIZE];
static Entry **storage;
static int storage_size = 0;
static int storage_index = 0;  // Index to insert new states

static int instantiate_index = 0; // Selected index for instantiating


/* Implementations of functions */

// Initializes the storage.
// Should be called only once at the start of the program.
void storage_init() {
	
	storage_size = 2;
	storage = (Entry **) calloc(storage_size, sizeof(Entry *));
	if (storage == NULL) {
		fprintf(stderr, "Error. Could not allocate memory for storage table. Program will abort.\n");
		exit(-1);
	}
	storage_index = 0;
	
	// Initialize mtype table.
	init_mtab();	
	
}

void realloc_storage() {
	Entry **tmp;
	
	if (storage_index >= storage_size) {
		storage_size = storage_size * 2;
		tmp = (Entry **) realloc(storage, storage_size * sizeof(Entry *));
		if (tmp == NULL) {
			fprintf(stderr, "Error. Could not allocate memory to increase storage table. Program will abort.\n");
			exit(-1);
		}
		storage = tmp;
		memset(&storage[storage_index], '\0', storage_size - storage_index);
	}
}

// Stores the BFS_Trail to be restored later (for expansion)
// Return value is currently not used
int save_frame(BFS_Trail *trail) {
	Entry *entr;
	Trail_Store *store;
	index_Store *ind;
	
	
	// Make sure there is enough space to allocate new entries.
	realloc_storage();		
	
	// Use calloc to allocate a place to store the state.
	// (Calloc initializes to 0, while malloc doesn't)
	storage[storage_index] = calloc(1, sizeof(Entry));
	entr = storage[storage_index];
	
	entr->store = calloc(1, sizeof(Trail_Store));
	store = entr->store;
	
	ind = new_index_store();
	
	store->trail = trail;
	if ( store->first_index == NULL ) {
		store->first_index = store->last_index = ind;	
	} else {
		store->last_index->next = ind;	
	}
	ind->index = storage_index;
	ind->next = NULL;
	
	push_index(&explore_cb_indices, storage_index);
	
	store->flags = 0; // Clean state
	store->hash_loc = Lstate; // Lstate was set globally by hstore()
	
	if (trans_text) {
		//fprintf(stderr, "A_LOG Transition: %s\n", trans_text);
		entr->trans_text = strdup(trans_text);
		trans_text = NULL;
		
		
		if (trans_vars) {
			//fprintf(stderr, "A_LOG Saw pointers to store...\n");
			//store->flags |= 2; // Set state to be instantiated
			entr->trans_vars = trans_vars;
			trans_vars = NULL;
		}
	} else if ( storage_index == 0 ) {
		entr->trans_text = strdup("");
	} 
#ifdef PREFER_END_TRANS
	else if (strncmp("-end-", trail->frame->o_t->tp, 5) == 0 ) {
		// We found an -end- transition. Save the index of it.
		explore_end_trans = storage_index;
	}
#endif

#ifdef SPIN_STATS
	if (storage_index % SPIN_MOD_STATES == 0) {
		fprintf(stderr, "A_LOG Number of Promela states: %.0f. Memory used by SPIN %-6.3f Mbyte\n", nstates, memcnt/1000000.);
	}
#endif
	
	storage_index += 1;

	return 0;
}

// now_tp and saved_tp may not be NULL. now_text and saved_text are allowed to be NULL
int state_label_eq(char *now_tp, char *now_text, char *saved_tp, char *saved_text) {
#ifdef	PROM_LABELS_MUST_MATCH
	if (now_text == NULL && saved_text == NULL && (strcmp(now_tp, saved_tp) == 0)) {
		return 1;
	} else if (now_text != NULL && saved_text != NULL && (strcmp(now_text, saved_text) == 0) ) {
		return 1;
	} else {
		return 0;
	}
//	if (now_text == NULL && saved_text != NULL) {
//		return 0;
//	} else if (now_text != NULL && saved_text == NULL) {
//		return 0;
//	} else if (now_text != NULL && (strcmp(now_text, saved_text) == 0) ) {
//		return 1;
//	} else if (now_text == NULL && (strcmp(now_tp, saved_tp) == 0) ) {
//		return 1;
//	} else {
//		return 0;
//	}
#else	
	if (now_text == NULL && saved_text == NULL) {
		return 1;
	} else if (now_text != NULL && saved_text != NULL && (strcmp(now_text, saved_text) == 0) ) {
		return 1;
	} else {
		return 0;
	}
#endif
}

int save_duplicate(Trail *tmp) {
	// Locate the duplicate state (by comparing the current Lstate
	// with the stored states ->hash_loc).
	int i;
	index_Store *ind;
	
	
	// TODO
	// If a state is found that matches hash_loc and label, create new index and append it to that state
	// If a state is found that matches only hash_loc, remember that state's 'store'. If no complete match
	// is found, at the end create a new Entry (and index), but link to the saved 'store', and declare that
	// it should not be saved by push_bfs (and save_frame)
	

	// Search backwards, because it (probably) is more likely to find a duplicate at the
	// end of the list
	for(i = storage_index - 1; i >= 0; i--) {
		// If the state is equal
		if ( 	storage[i] != NULL && 
			storage[i]->store->hash_loc == Lstate &&
			storage[i]->store->trail->frame->o_t != NULL && // Make sure state has source label. Initial state doesn't.
			state_label_eq(tmp->o_t->tp, trans_text, storage[i]->store->trail->frame->o_t->tp, storage[i]->trans_text) 
		   ) {
#ifdef PREFER_END_TRANS
		   	if (strncmp("-end-", tmp->o_t->tp, 5) == 0 ) {
				// We found an -end- transition. Save the index of it.
				explore_end_trans = (explore_known) ? i: storage_index;
			}
#endif			
		   	if (explore_known) {
		   		// If we know we explore a known state, we re-use the previous index.
			   	push_index(&explore_cb_indices, i);
				free(trans_text);
				trans_text = NULL;
				free_var_stores(&trans_vars);
				trans_vars = NULL;
				// In this case pan will consider the state to be a duplicate and not save it
			   	return 1;
			} else {
				// Make sure there is enough space to allocate new entries.
				realloc_storage();		

				// Store the entry, at a new index.
				storage[storage_index] = calloc(1, sizeof(Entry));
				storage[storage_index]->trans_text = trans_text;
				storage[storage_index]->trans_vars = trans_vars;
				storage[storage_index]->store = storage[i]->store;
				trans_vars = NULL;
				trans_text = NULL;
					
				push_index(&explore_cb_indices, storage_index);
				
				// Append index at end of indices of state.
				ind = new_index_store();
				storage[i]->store->last_index->next = ind;
				storage[i]->store->last_index = ind;
				ind->index = storage_index;
				ind->next = NULL;

#ifdef SPIN_STATS
	if (storage_index % SPIN_MOD_STATES == 0) {
		fprintf(stderr, "A_LOG Number of Promela states: %.0f. Memory used by SPIN %-6.3f Mbyte\n", nstates, memcnt/1000000.);
	}
#endif
				
				storage_index += 1;
				
				// In this case pan will consider the state to be a duplicate and not save it
				return 1;
			}
		   }
	}
	fprintf(stderr, "A_LOG Did not find duplicate state with same label. Storing as new state instead.\n");
	// In this case, pan will consider the state as 'new' and save it (even though it has a duplicate Lstate)
	return 0;
				
	 
}


// Return values for cmd_* functions.
// INDEX_QUIT is used to indicate the program has to end 
#define INDEX_QUIT  -1
// INDEX_RETRY is used to indicate that the (explorer_)interface should continue to read commands
#define INDEX_RETRY -2

// Used to select a frame (through interaction) for expansion. Uses either a function for human
// interaction, or an explorer_interface function.
// Returns the selected frame to be expanded
BFS_Trail *load_frame() {
	int index = INDEX_RETRY;
	
	// Shortcut for instantiating a state
	if ( instantiate_index > 0 ) {
		return (BFS_Trail *) storage[instantiate_index]->store->trail;
	}	

	while (index == INDEX_RETRY) {
		#ifdef SPEX_INTERACT
		index = interact();
		#else
		index = explorer_interface();
		#endif
	}
	// while continues as long as:
	// a) the index returned is INDEX_RETRY _or_
	// b) the index is valid, but there is no state stored there _or_
	// c) the index is valid, but the state has already been expanded
	
	if (index == INDEX_QUIT) {
		return NULL;	
	} else {
		storage[index]->store->flags |= 1; // Set dirty bit on frame, to indicate it has been expanded before
		return (BFS_Trail *) storage[index]->store->trail;
	}
}

index_Store *new_index_store() {
	index_Store *ind;
	
	ind = calloc(1, sizeof(index_Store));
	if (ind == NULL) {
		printf("A_ERROR Unable to allocate new index storage. Program will exit now.\n");
		exit(-1);
	}
	return ind;
}

void free_index_store(index_Store *ind) {
	free(ind);
}

index_Store *push_index(index_Store **stack, int index) {
	index_Store *ind = NULL;
	
	if (index > 0 ) {
		ind = new_index_store();
		ind->index = index;
		ind->next = *stack;
		*stack = ind;
	}
	return ind;		
}

void free_indices(index_Store **stack) {
	index_Store *ind;
	index_Store *ind_next;
	
	// Remove all indices that were stored for callback.
	for (ind = *stack; ind != NULL; ind = ind_next) {
		ind_next = ind->next;
		free_index_store(ind);
	}
	*stack = NULL;	
	
}

char *print_state(int event) {
	Var_Store *vstore = NULL;
	static char *tmp;
	static int tmp_length;
	int tmp_index = 0;
	
	//int i;
	char *s, *e;
	
	char charString[] = { '\0', '\0' };
	
	// Ee TAB event TAB visible TAB solved TAB label TAB preds TAB freevars TAB identical
	
	char event_text[10] = ""; // String version of label.

	char visible = 0;

	//char solved = 1;

	static char *label;
	static int label_length;
	int label_index = 0;
	
	static char *preds;
	static int preds_length;
	int preds_index = 0;
	
	static char *freevars = 0;
	static int freevars_length = 0;
	int freevars_index = 0;

	char identical[10] = ""; // To be set later
	// The space needed for 'identical' depends on the amount of
	// states that can be stored. 9 digits mean we can print up
	// to 999,999,999 states. (Which seems sufficient.)
	
	if (storage[event] == NULL ) {
		return NULL;
	}
	
	sprintf(event_text, "%d", event);
	
	// set visibility
	visible = (storage[event]->trans_text != NULL);
	

	if (freevars) {
		freevars[0] = '\0';
	}
	if (label) {
		label[0] = '\0';
	}
	
	// set label
	if (storage[event]->trans_text) {
		// Normalize variables here.
		// set free variables
		s = storage[event]->trans_text;
		e = s + strlen(s);

		fprintf(stderr, "A_DEBUG trans_text for %d is: |%s|\n", event, s);
		for(;s < e; s++) {
			if ( *s == '%' ) {
				s++; 
				if ( vstore == NULL ) {
					vstore = storage[event]->trans_vars;
				} else {
					vstore = vstore->next;
				}
				incr_strcat(&freevars, &freevars_index, &freevars_length, vstore->norm_var_name, 15);
				incr_strcat(&freevars, &freevars_index, &freevars_length, " ", 15);
				incr_strcat(&freevars, &freevars_index, &freevars_length, vstore->org_var_name, 15);
				incr_strcat(&freevars, &freevars_index, &freevars_length, " ", 15);
				incr_strcat(&freevars, &freevars_index, &freevars_length, vstore->type, 15);
				incr_strcat(&freevars, &freevars_index, &freevars_length, " ", 15);

				incr_strcat(&label, &label_index, &label_length, vstore->norm_var_name, 15);
			} else {
				charString[0] = *s;
				incr_strcat(&label, &label_index, &label_length, charString, 15);
			}
		}
	} else {
		// Use transmognify function from pan.c to print "pretty" labels (even for c_code)
		label = incr_strcat(&label, &label_index, &label_length, transmognify(storage[event]->store->trail->frame->o_t->tp), 15);
	}
	
	//fprintf(stderr, "A_DEBUG label was: |%s|\n", label);
	
	// Set predicates string.
	// This will contain the range for each (normalized) variable in the form:
	// "norm_var1_name in { range1 }, norm_var2_name in { range2 }"
	preds_index = 0;
	
	if (preds) {
		preds[0] = '\0';
	}
	if (storage[event]->trans_vars) {
				
		for ( vstore = storage[event]->trans_vars; vstore != NULL; vstore = vstore->next) {
			// TODO Print only domains that are different than the default.
			// Also, do not print mtypes.
			incr_strcat(&preds, &preds_index, &preds_length, vstore->norm_var_name, 10);
			incr_strcat(&preds, &preds_index, &preds_length, " in ", 10);
			incr_strcat(&preds, &preds_index, &preds_length, DomainToString(vstore->var_domain), 10);			
			
			if (vstore->next != NULL) {
				incr_strcat(&preds, &preds_index, &preds_length, "; ", 10);
			}
				
		}
		
		
	}
			
	
	// set identical index
	if (storage[event]->store->first_index->index != event) {
		sprintf(identical, "%d", storage[event]->store->first_index->index);
	} else {
		identical[0] = '\0';
	}
			
	// Ee TAB event TAB visible TAB solved TAB label TAB preds TAB freevars TAB identical
	// sprintf(tmp, "%s\t%d\t%d\t%s\t%s\t%s\t%s", event_text, visible, solved, label, preds, freevars, identical);
	incr_strcat(&tmp, &tmp_index, &tmp_length, event_text, 15);
	incr_strcat(&tmp, &tmp_index, &tmp_length, "\t", 15);
	incr_strcat(&tmp, &tmp_index, &tmp_length, visible ? "1" : "0", 15);
	incr_strcat(&tmp, &tmp_index, &tmp_length, "\t1\t", 15); // solved is always 1
	incr_strcat(&tmp, &tmp_index, &tmp_length, label, 15);
	incr_strcat(&tmp, &tmp_index, &tmp_length, "\t", 15);
	incr_strcat(&tmp, &tmp_index, &tmp_length, preds, 15);
	incr_strcat(&tmp, &tmp_index, &tmp_length, "\t", 15);
	incr_strcat(&tmp, &tmp_index, &tmp_length, freevars, 15);
	incr_strcat(&tmp, &tmp_index, &tmp_length, "\t", 15);
	incr_strcat(&tmp, &tmp_index, &tmp_length, identical, 15);
	
	return strdup(tmp);	
}


void delete_state(int state_index) {

	index_Store *ind;
	index_Store *ind_next;
	
	// TODO Fix deleting states.
	
	fprintf(stderr, "deleting of states has been temporarily disabled.\n");	
	return;
/*	
	fprintf(stderr, "delete_state has not been updated. This code is untested and should (probably) not be used.\n");

	if (storage[state_index] != NULL && storage[state_index]->store->first_index->index == state_index) {
		// We remove the first index of this state. Search it's list of indices
		// for a new valid first index. This will delete any old indices before
		// the new first index.	
		
		ind = storage[state_index]->store->first_index;
		ind_next = ind->next;
		free_index_store(ind);
		ind = ind_next;
		
		while (ind && storage[ind->index] && storage[state_index] && storage[ind->index]->store != storage[state_index]->store) {
			ind_next = ind->next;	
			free_index_store(ind);
			ind = ind_next;
		}
		
		if ( ind == NULL ) {
			// There were no valid indices left for this state, so we know
			// we can remove it safely.
			
			delete_trail(storage[state_index]->store, state_index);
			// TODO delete_entry(storage[state_index], state_index);
		} else {
			// We found a new valid first index at 'ind'
			storage[state_index]->store->first_index = ind;
		}	
	}
	storage[state_index] = NULL;
*/	
		
}

void delete_trail(Trail_Store* store, int store_index) {
	// Function to delete the state stored in "store".
	// Use "freesv(SV_Hold *)" to free the SV_Hold elements, 
	// and add the BFS_Trail elements to the "bfs_free" linked list
	// Also should "free()" the Trail_Store element (if valid)

	if (store != NULL) {
		if ((store->flags&1) == 0) {
			fprintf(stderr, "A_WARNING Deleting unexpanded state %d.\n", store_index);
		}
		freesv(store->trail->onow);
		
		store->trail->nxt = bfs_free;
		bfs_free = store->trail;
		
		// TODO !! Should remove state from PAN hashtable as well !!
		
		free_trail(store);
	}
}

void free_trail(Trail_Store* store) {
	// TODO Put in free_entry(): free(store->trans_text);
	free(store);
}

/* *********************************************************************************** */
/* *********************************************************************************** */
/* ************************* Begin of interaction functions ************************** */
/* *********************************************************************************** */
/* *********************************************************************************** */



/* Interact function is depracated.

// Define line as global, to prevent constant loading/discarding from stack.
// Only to be used (in global context) by the interact() and 
// explorer_interface() function. 
//static char line[MAX_LINE_LENGTH];

// Human interaction function to select a frame for expansion
// Returns the index of the frame to be expanded
// Depracated
int interact() {
	int index;
	int scanned_args;
	char *got;
	
	printf("Which frame would you like to load?\n");
	printf("You can choose from frames between 0 and %d.\n", storage_index - 1);
	printf("\nI choose: ");
	
	got = fgets(line, MAX_LINE_LENGTH, stdin);
	if (got == NULL) {
		printf("Error while reading line. Program will quit now.\n");
		return INDEX_QUIT;
	} else {	
		scanned_args = sscanf(got, "%d", &index);
		if (scanned_args != 1) {
			if (got[0] == 'q' || got[0] == 'x') {
				return INDEX_QUIT;
			} else {
				// if we see something we don't recognize, just return INDEX_RETRY
				printf("Invalid input. Try again. (Use 'q' to quit.)\n");
				return INDEX_RETRY;
		  	}	
		} else if ( index > storage_index - 1 || index < 0 ) {
	  		printf("Index is out of bounds. Try again.\n");
	  		return INDEX_RETRY;
		} else if (storage[index] == NULL) {
			printf("Index is invalid. Try again.\n");
			return INDEX_RETRY;
		} else if ( storage[index]->store->flags&1) {
			printf("State has been expanded before. Try again.\n");
			return INDEX_RETRY;
	  	} else {
	  		return index;
	  	}
	}
}

*/ // end of interact function

// Function that implements the explorer interface as defined for TorX. It is called every time
// a new frame has to be selected for expansion (the main function of the tool after all).
// If an explore step has been done before this, it will call the explore_callback function
// so it can print the newly generated states.
// It reads a line from stdin, which will be handled by the handle_line function.
// Returns the index of the frame that is selected for expanding.
int explorer_interface() {
	// This function will handle the reading of input, then delegate the
	// read line to the handle_line() function.
	
	static char *in_buf;
	static int in_buf_len;

	
	if (explore_cb) {
		explore_callback();
	}

	fflush(stderr);
	fflush(stdout);
	
	if( incr_fgets(&in_buf, &in_buf_len, stdin, 20) != NULL ) {
		return handle_line(in_buf);
	} else {
		// Something went wrong while trying to read the line...
		fprintf(stderr, "A_ERROR Unable to read input (EOF reached). Program will quit now.\n");
		return INDEX_QUIT;
	}
}

// Selects the right cmd_* function, based on the first character of the input
// Returns the index of the frame to be expanded.
int handle_line(char *input) {
	char *parse_input;
	
	fprintf(stderr, "A_DEBUG Command received: %s", input);
	
	for ( parse_input = input; isspace(*parse_input); parse_input++ ) {
		;
	}
	switch (parse_input[0]) {
		default:  return cmd_unknown(parse_input);
		case 'q': return cmd_quit(parse_input);
		case 'r': return cmd_reset(parse_input);
		case 'e': return cmd_explore(parse_input);
		case 'd': return cmd_delete(parse_input);
		case 'i': return cmd_instantiate(parse_input);
	}	
}

 
int cmd_unknown(const char *input) {
	// No ending '\n' here, because the input string already has one
	fprintf(stderr, "A_ERROR Unknown command: %s", input);
	printf("%c0 %s", toupper(input[0]), input+1);
	return INDEX_RETRY;
}

int cmd_quit(const char *input) {
	printf("Q\n");
	return INDEX_QUIT;
}

int cmd_reset(const char *input) {
	// During a reset all states (except initial) should be deleted
	// Use the "delete_state(int)" function for this.
	int i;
	
	for (i = storage_index; i >= 1; i--) {
		delete_state(i);
	}
	
	
	storage_index = 1;
	//storage_oindex = 1;
	
	storage[0]->store->flags = 0;
	
	// 0 for initial state, 1 for solved
	printf("R\t%d\t%d\n", 0, 1);
	return INDEX_RETRY;
}

int cmd_explore(const char *input) {
	int select;
	int scanned;
	
	explore_cb = 0;
	// "input" contains the whole string read from the commandline
	// This includes the leading 'e', which has to be skipped
	input += 1;
	
	scanned = sscanf(input, "%d", &select);
	if (scanned == 0) {
		printf("E0 ParseError: not a number: %s\n", input);
		select = INDEX_RETRY;
	} else if (select < 0 || select >= storage_index || storage[select] == NULL || storage[select]->store->trail == NULL) {
		printf("E0 Unknown event number: %d\n", select);
		select = INDEX_RETRY;
	} else if (storage[select]->trans_vars) {
		// If the state has to be instantiated, print and retry.
		printf("E0 State has to be instantiated before expanding.\n");
		select = INDEX_RETRY;
	} else {
		// We've selected a valid state
		// Set to callback after we're done expanding this state
		explore_cb = 1;
		if (storage[select]->store->flags&1) {
			// If the state was expanded before, indicate so (adjusts behaviour of save_duplicate)
			fprintf(stderr, "A_DEBUG State has been expanded before. (Shortcutting 'save_duplicate()'.)\n");
			explore_known = 1;
		} 
	}
	return select;
}

void explore_callback(void) {
	// Here we can print some stuff if the explore command
	// found new states
	
	index_Store *ind;
	char *state_answer;
	
	char found_new = 0;
	
	for (ind = explore_cb_indices; ind != NULL; ind = ind->next) {
		if (found_new == 0) {
			printf("EB\n");
			found_new = 1;
		}
		// TODO
#ifdef PREFER_END_TRANS
		if (explore_end_trans == 0 || ind->index == explore_end_trans) 
#endif
		{
			state_answer = print_state(ind->index);
			printf("Ee\t%s\n", state_answer);
			free(state_answer);
		}
	}
	if (found_new != 0) {
		printf("EE\n");
	} else {
		printf("E0 No new states found\n");
	}
	
	free_indices(&explore_cb_indices);

	explore_known = 0;
	explore_cb = 0;
#ifdef PREFER_END_TRANS	
	explore_end_trans = 0;
#endif
}

int min (int a, int b) {
	if ( a < b )
		return a;
	else 
		return b;
}

int cmd_instantiate(const char *input) {
	Trail *t;
	index_Store *ind;
	Var_Store *vstore = NULL;
	int select;
	int scanned;
	char *str;
	int length;
	int i;
	
	static char *new_trans_text;
	static int new_trans_length;
	int new_trans_index = 0;
	
	 // need to hold longest value representation which right now
	 // is -MIN_INT which takes 1(-)+10(digits)+1(\0) = 12
	char tmp_string[15];
	
	//char *temp_trans_cursor = temp_trans_text;
	//int temp_trans_length = 0;
	
	char expr[80] = { '\0' };
	int expr_length = 0;
	long expr_value = 0;
	char *expr_ind_1;	// Beginning of expression
	char *expr_ind_2 = 0;	// Index of first '=' after expression
	char *expr_ind_3 = 0;	// Index of first ';' after expression
	
	int *int_pointer;
	//char *char_pointer;
	//short *short_pointer;
	
	memset(tmp_string, '\0', sizeof(tmp_string));
	
	//incr_strcat(&new_trans_text, &new_trans_index, &new_trans_length, " ", 5);
	
	// "input" contains the whole string read from the commandline
	// This includes the leading 'i', which has to be skipped
	input += 1;
	
	scanned = sscanf(input, "%d", &select);
	if (scanned == 0) {
		printf("I0 ParseError: not a number: %s\n", input);
		return INDEX_RETRY;
	} else if (select < 0 || select >= storage_index || storage[select] == NULL || storage[select]->store->trail == NULL) {
		printf("I0 Unknown event number: %d\n", select);
		return INDEX_RETRY;
	} else {
		// We've selected a valid state
		// If the state does not have to be instantiated, print and retry.
		if (! (storage[select]->trans_vars)) {
			printf("I0 State cannot be instantiated.\n");
			return INDEX_RETRY;
		}
	}
	
	// Set global instantiate_index to shortcut the load_state function, so it will load the 'select'ed state.
	instantiate_index = select;
	
	// Pop state so it is loaded in memory.
	t = pop_bfs();


	
	/*
Typename	C-equivalent	Macro in limits.h	Typical Range
bit or bool	bit-field	-			0..1
byte		uchar		CHAR_BIT width in bits  0..255
short		short		SHRT_MIN..SHRT_MAX	-2^15 - 1 .. 2^15 - 1
int		int		INT_MIN..INT_MAX	-2^31 - 1 .. 2^31 - 1
	*/

		
	str = storage[select]->trans_text;
	//fprintf(stderr, "trans_text = \"%s\"\n", str);
	length = strlen(str);
	for ( i = 0; i < length; i++ ) {
		if ( str[i] == '%' ) {
			if ( vstore == NULL ) {
				vstore = storage[select]->trans_vars;
			} else {
				vstore = vstore->next;
			}
			
			if (vstore == NULL) {
				fprintf(stderr, "A_ERROR Not enough variables stored for number of '%%' in trans_text\n");
				return INDEX_QUIT;
			}
			
			expr_ind_1 = strstr(input, vstore->org_var_name); 
			
			// If we don't find the normalized name of this variable in the input line,
			// we should append it to the list of variables that will still need to be
			// instantiated. After this, we continue with the next variable.
			if (expr_ind_1 == NULL) {
				append_var_store(dup_var_store(vstore));
				// Copy the next two characters (the '%', followed by type)
				sprintf(tmp_string, "%.2s", &str[i]);
				incr_strcat(&new_trans_text, &new_trans_index, &new_trans_length, tmp_string, 5);

				i += 1;			
				continue; 
			}
			//fprintf(stderr, "expr_ind_1 = \"%s\"\n", expr_ind_1);
			expr_ind_2 = strstr(expr_ind_1, "=");
			if ( expr_ind_2 == NULL ) {
				printf("I0 No '=' found in expression.\n");

				trans_text = NULL;
				free_var_stores(&trans_vars);
				trans_vars = NULL;
				return INDEX_RETRY;
			}
			// Skip over '=' sign.
			expr_ind_2++;
			
			// Locate end of expression (either ';' or '\n'), whichever comes first :)
			// This might allow someone to enter an expression of "var1 = 1 var2 = 2 var3 = 3"
			// which will be allowed (and work correctly)
			expr_ind_3 = strpbrk(expr_ind_1, ";\n");
			if ( expr_ind_3 == NULL ) {
				printf("I0 Could not locate end of expression\n");

				trans_text = NULL;
				free_var_stores(&trans_vars);
				trans_vars = NULL;
				return INDEX_RETRY;
			}

			while ( isspace(*expr_ind_2) && expr_ind_2 <= expr_ind_3) {
				expr_ind_2++;
			}

			//"check if maybe last argument"
			if (expr_ind_2 >= expr_ind_3) {
				i++;
				continue;
			} 
			
			
			// We have found an expression for the free variable, prepare it.
			expr_length = min(expr_ind_3 - expr_ind_2, 79);
			strncpy(expr, expr_ind_2, expr_length);
			expr[expr_length] = '\0';
			
			
			// Convert expression to long value. After this, we can cast it to the type we need.
			switch (str[i+1]) {
				case 'd': // Fall through
				case 'c': // Fall through
				case 'h': {
					// Base 0 means either base 8, 10 or 16, depending on leading chars.
					expr_value = strtol(expr, NULL, 0);
					break;
				}
				case 'e': {
					expr_value = mfind(expr);
					break;
				}
				default: {
					fprintf(stderr, "Invalid field (%c) specified in label. Field removed.\n", str[i+1]);
					i += 1;
					continue;
				} 
			}
			//fprintf(stderr, "A_LOG %s = |%s| (%ld)\n", vstore->norm_var_name, expr, expr_value);
			
			if ( str[i+1] == 'e' && expr_value < 1 ) {
				printf("I0 The specified mtype (%s) is invalid.\n", expr);
				trans_text = NULL;
				free_var_stores(&trans_vars);
				trans_vars = NULL;
				return INDEX_RETRY;
			}				
			
			if (! DHas(vstore->var_domain, (RANGE_VALUE_TYPE) expr_value) ) {
				printf("I0 Value %d is not fit for the domain %s.\n", expr_value, vstore->norm_var_name);

				trans_text = NULL;
				free_var_stores(&trans_vars);
				trans_vars = NULL;
				return INDEX_RETRY;
			}

			// Look for the next character to see what kind of variable we need.
			switch (str[i+1]) {
				case 'd': {
					*vstore->pointer.i = (int) expr_value;
					sprintf(tmp_string, "%d", *vstore->pointer.i);
					break;
				}
				case 'c': {
					*vstore->pointer.c = (unsigned char) expr_value;
					sprintf(tmp_string, "%u", (unsigned char) *vstore->pointer.c);
					break;
				}
				case 'h': {
					*vstore->pointer.s = (short) expr_value;
					sprintf(tmp_string, "%hd", *vstore->pointer.s);
					break;
				}
				case 'e': {
					*vstore->pointer.c = (char) expr_value;
					tmp_string[0] = '\0';
					incr_strcat(&new_trans_text, &new_trans_index, &new_trans_length, mstring((char) expr_value), 10);
				}				
					
			}
			incr_strcat(&new_trans_text, &new_trans_index, &new_trans_length, tmp_string, 10);				

			// Increase i to skip indicated variable.
			i++;
		} else {
			tmp_string[0] = str[i];
			tmp_string[1] = '\0';
			incr_strcat(&new_trans_text, &new_trans_index, &new_trans_length, tmp_string, 5);

//			*temp_trans_cursor = str[i];
//			temp_trans_cursor += 1;
		}
	}

//	*temp_trans_cursor = '\0';
	trans_text = strdup(new_trans_text);
	//fprintf(stderr, "trans_text = %s\n", trans_text);

	
	store_state(t, 1, boq);
	
	ind = explore_cb_indices;
	//fprintf(stderr, "A_LOG Instantiated to state %d\n", ind->index);
	
	
	str = print_state(ind->index);
	printf("I\t%s\n", str);
	free(str);	

	free_indices(&explore_cb_indices);
	instantiate_index = 0;
	trans_vars = NULL;
	trans_text = NULL;

	return INDEX_RETRY;	
}


int cmd_delete(char *input) {
	// TODO Needs to be updated for Entries
	char *select;
	int selected;
	static char *whitespace = " \t\n\r";
		
	// First call of 'strtok' will remove the 'd' in front of the input
	strtok(input, whitespace);
	
	// Next calls of 'strtok' will iterate over the indices in the input
	while ( (select = strtok(NULL, whitespace)) != NULL ) {
		selected = atoi(select);
		if (selected != 0) {
			fprintf(stderr, "A_LOG Deleting state %s (%d)\n", select, selected);
			delete_state(selected);
		}
	}

	printf("D\n");
	
	return INDEX_RETRY;
}


// End of file
