# ===============================================================

# Application specific config file processing code.

proc config_init {} {
	global config
	
	debugMsg "config_init"
	catch {unset config}
	set config(emset) [em_init]
	set config(channels) {}
}
proc config_finish {} {
	global config
	
	em_compile $config(emset)
}
proc config_cleanup {} {
	catch {unset config}
}
proc config_defaults {iokind} {
	global primer
	switch -- $iokind {
		input {
		} output {
		    switch -- $primer(algorithm) {
		    	ioco -
		    	ioct -
		    	sim {
		    		return {
						if {[string compare $suspension ""] == 0} {
							set suspension delta
						}
						if {[string compare $suspmode ""] == 0} {
							set suspmode compute
						}
					}
		    	} traces {
		    		return {
						if {[string compare $suspension ""] == 0} {
							set suspension delta
						}
						if {[string compare $suspmode ""] == 0} {
							set suspmode recognize
						}
					}
				} default {
					warningMsg "ConfigError: config_defaults: unknown primer algorithm: $primer(algorithm)"
				}
		   	}
		} default {
			warningMsg "ConfigError: config_defaults: illegal value: iokind should be in (input,output): $iokind"
		}
	}
	return {}
}

proc config_process_fields {status_fields} {
	global config
	debugMsg "Status-Fields: $status_fields"
	if {![lindex $status_fields 0]} {
		return
	}
	set fields [lindex $status_fields 1]
	debugMsg "Fields: $fields"
	foreach name {channel verdict iokind suspension suspmode} {
		set $name [field_value $fields $name]
	}
	foreach name {event} {
		set $name [field_value_array $fields $name]
	}
	if {([string compare $channel ""] != 0) &&
	    ([string compare $verdict ""] != 0)} {
		warningMsg "ConfigInconsistency: only one allowed of channel, verdict"
	}
	if {[llength $event] <= 0} {
		warningMsg "ConfigError: field missing: event"
	}
	if {[string compare $channel ""] != 0} {
		if {[lsearch $config(channels) $channel] >= 0} {
			warningMsg "ConfigError: duplicate configuration for channel: $channel"
		} else {
			lappend config(channels) $channel
		}
		switch -- $iokind {
			 "" {
				warningMsg "ConfigError: field missing: iokind"
			} input {
		    	eval [config_defaults $iokind]
		    } output {
		    	eval [config_defaults $iokind]
		    } default {
				warningMsg "ConfigError: illegal value: iokind should be in (input,output): $iokind"
			}
		}
		lappend config(channels,$iokind) $channel
		set config(iokind,$channel) $iokind
		foreach p $event {
			em_add_pattern $config(emset) $p [list $iokind $channel]
		}
		if {[string compare $suspension ""] != 0} {
			if {[string compare $suspmode recognize] != 0 && [regexp {\*} $suspension]} {
				warningMsg "ConfigError: suspension not allowed to contain wildcards: $suspension"
			} else {
				lappend config(channels,[set iokind]suspension) $channel
				em_add_pattern $config(emset) $suspension [list [set iokind]suspension $channel]
				set config(suspension,repr,$iokind,$channel) $suspension
			}
		}
		set config(suspension,detect,$channel) 0
		switch -- $suspmode {
			  "" {
			} compute {
				set config(suspension,detect,$channel) 1
			} recognize {
				set config(suspension,detect,$channel) 0
			} default {
				warningMsg "ConfigError: illegal value: suspmode should be in (compute,recognize): $suspmode"
			}
		}
	} elseif {[string compare $verdict ""] != 0} {
		foreach p $event {
			em_add_pattern $config(emset) $p [list verdict $verdict]
		}
		
	} else {
		warningMsg "Config: missing: config block should contain channel or verdict"
	}
}

proc config_get_channels {iokind} {
	global config
	if {[info exists config(channels,$iokind)]} {
		return $config(channels,$iokind)
	} else {
		return {}
	}
}

proc config_get_iokind {chan} {
	global config
	if {[info exists config(iokind,$chan)]} {
		return $config(iokind,$chan)
	} else {
		return ""
	}
}

proc config_get_suspension_representation {channel} {
	global config
	set res ""
	if {[info exists config(iokind,$channel)]} {
		set iokind $config(iokind,$channel)
		if {[info exists config(suspension,repr,$iokind,$channel)]} {
			set res $config(suspension,repr,$iokind,$channel)
		}
	}
	return $res
}

proc config_suspension_detection {channel} {
	global config
	if {[info exists config(suspension,detect,$channel)]} {
		return $config(suspension,detect,$channel)
	} else {
		return 0
	}
}

proc config_is_channel { chan } {
	global config
	if {[info exists config(iokind,$chan)]} {
		return 1
	} else {
		return 0
	}
}

proc config_is_iokind { kind } {
	global config
	if {[info exists config(channels,$kind)]} {
		return 1
	} else {
		return 0
	}
}

# ===============================================================

# General (application independent) config file processing code.

proc config_from_file {filename} {
	if {[catch {open $filename r} fileId]} {
		warningMsg "Cannot open $filename: $fileId"
	} else {
		config_process_file $fileId
		close $fileId
	}
}

proc config_process_file {channel} {
	foreach line [split [read $channel] \n] {
		if {[config_is_whitespace_or_comment $line]} {
			continue
		}
		if {[config_is_start_of_block $line]} {
			# handle fields of previous block, if any
			config_process_fields [config_list_collected_fields]
			config_clear_collected_fields
		}
		config_collect_fields [fields_split $line " \t"]
	}
	# handle fields of previous (last) block, if any
	config_process_fields [config_list_collected_fields]
	config_clear_collected_fields
}

proc config_is_whitespace_or_comment {line} {
	if {[regexp {^\s*(#.*)?$} $line]} {
		return 1
	} else {
		return 0
	}
}

proc config_is_start_of_block {line} {
	if {[regexp {^[^ \t]} $line]} {       # start of new block
		return 1
	} else {
		return 0
	}
}

proc config_collect_fields {fields} {
	global configtmpfields
	if {![info exists configtmpfields]} {
		set configtmpfields {}
	}
	# append fields to 'current' block
	set configtmpfields [concat $configtmpfields $fields]
}

proc config_list_collected_fields {} {
	global configtmpfields
	if {![info exists configtmpfields]} {
		return [list 0 {}]
	} else {
		return [list 1 $configtmpfields]
	}
}

proc config_clear_collected_fields {} {
	global configtmpfields
	catch {unset configtmpfields}
}

# ===============================================================
