# =========================================================================
# $Id: menurepr.tcl,v 1.1.2.12 2007/07/03 10:38:56 belinfan Exp $
# =========================================================================

## Build a menu structure, containing the 'current' menu of events.

## Each event is represented by a tuple containg:
##	- event number (nat)
##	- visible (0 = internal, 1 = external)
##	- preds_solved (0/1)
##	- label (with normalised free variable names)
##	- predicates (with normalised free variable names)
##  - free_vars list of 3-tuples <org-name,type,normalised-name>?)
##	- identical (to which state is the state reached by this event identical to?)
## When an event is inserted, together with the event a 'kind' is given
## (input output inputsuspension outputsuspension internal verdict)
## together with a channel (for MIOCO).

## The menu is represented by a list of lists, where each sublist
## contains menu-elements with 'identical' events, i.e.
## if one event of the list is 'done', all other events of the list
## can be, and are, 'done' (followed, whatever) as well.
## All events in the same sublist have the same
##  - kind 
##	- label (with normalised free variable names)
##	- predicates (with normalised free variable names)
## Each of the events in the sublist has its own
##	- event number (nat)
##	- preds_solved (0/1)
##  - free_vars list of 3-tuples <org-name,type,normalised-name>?)
##	- identical (to which state is the state reached by this event identical to?)

## We try to optimise the list of lists, such that we do not put
## redundant events in a menu (sub)list. Which events are redundant?
## Redundant events are those for which an event with the same
## <kind,channel,label,predicates> is already in the menu,
## and also the <identical> of the redundant event is in the _same_
## <kind,channel,label,predicates> sublist.
## Things may get slightly tricky here, because we may have a
## situation in which the <identical> of an event is not in its
## <kind,channel,label,predicates> sublist, whereas this sublist
## _does_ contain an element with the same <identical>.
## Example:
## we want to insert <input,in,<10,1,1,a,,,4>> in sublist <input,in,a,>
## and this sublist does not contain an element <input,in,<4,1,?,a,,?,?>>
## then our first thought is: the event is not redundant.
## However, if the sublist does contain <input,in,<8,1,1,a,,,4>> then,
## indeed, event nr. 10 _is_ redundant.

## In the implementation, each sublist is identified by a number.
## These numbers are not globally unique, but only for each <kind,channel>.
## With each <kind,channel> we associate a list of sublist numbers -
## for internal, suspension and verdict events we have just one sublist;
## for each visible <kind,channel> we have a sublist for each
## <kind,channel,label,predicates> in the same <kind,channel>.

## The interface to this structure is formed by the
## a the following routines.
##
## Currently implemented actions:
proc mr_init {} {
	global mr
	catch {
		unset mr
	}
	set mr(kinds) [list input output inputsuspension outputsuspension internal verdict]
	foreach k [concat $mr(kinds)] {
		set mr(channels,$k) [get_channels $k]
		if {[llength $mr(channels,$k)] <= 0} {
			set mr(channels,$k) [list ""]
		}
		foreach c $mr(channels,$k) {
			set mr(keys,$k,$c) {}
			set mr(nextkey,$k,$c) 1
		}
	}
}

proc mr_empty {} {
	mr_init
}

# returns 1 when event has been inserted in menu
# returns 0 when event has not been inserted (identical/error)
proc mr_insert { kind chan etuple repr {attr {}} } {
	global mr
	debugMsg "mr_insert ($kind) ($chan) ($etuple) ($repr)"
	set event [lindex $etuple 0]
	set visible [lindex $etuple 1]
	set preds_solved [lindex $etuple 2]
	set label [lindex $etuple 3]
	set preds [lindex $etuple 4]
	set free_vars [lindex $etuple 5]
	set identical [lindex $etuple 6]
	return [mr_do_insert $kind $chan $repr $preds $event $etuple $attr]
}

proc mr_events { {kind_opt ""} } {
	global mr
	if {[string compare $kind_opt ""] != 0} {
		return [mr_do_events [list $kind_opt]]
	} else {
		return [mr_do_events $mr(kinds)]
	}
}

proc mr_show {} {
	return [join [mr_list] "\n"]
}
proc mr_display {} {
	return [mr_show]
}
proc mr_list { {kinds_opt {}} {chan_opt {}} } {
	global mr
	set events {}
	if {[llength $kinds_opt] > 0} {
		set kinds {}
		foreach kind $mr(kinds) {
			if {[lsearch -exact $kinds_opt $kind] >= 0} {
				lappend kinds $kind
			}
		}
	} else {
		set kinds $mr(kinds)
	}
	foreach kind $kinds {
		if {[llength $chan_opt] > 0} {
			set channels {}
			foreach chan $mr(channels,$kind) {
				if {[lsearch -exact $chan_opt $chan] >= 0} {
					lappend channels $chan
				}
			}
		} else {
			set channels $mr(channels,$kind)
		}
		foreach c $channels {
			set events [concat $events [mr_list_kind_chan $kind $c]]
		}
	}
	return $events
}


proc mr_free_variables {kind chan ev} {
	global mr
	if {[info exists mr(freevars,$kind,$chan,$ev)]} {
		return $mr(freevars,$kind,$chan,$ev)
	} else {
		return {}
	}
}

proc mr_repr {kind chan ev} {
	global mr
	if {[info exists mr(etuple,$kind,$chan,$ev)]} {
		return [lindex $mr(etuple,$kind,$chan,$ev) 3]
	} else {
		return ""
	}
}

proc mr_event {kind chan ev} {
	global mr
	if {[info exists mr(etuple,$kind,$chan,$ev)]} {
		return $mr(etuple,$kind,$chan,$ev)
	} else {
		return ""
	}
}

proc mr_identical {ev} {
	global mr
	if {[info exists mr(identical,$ev)]} {
		return $mr(identical,$ev)
	} else {
		return ""
	}
}

proc mr_get {} {
	global mr
	return [array get mr]
}
proc mr_set {l} {
	global mr
	return [array set mr $l]
}


# ===========================================================

proc mr_do_insert { kind chan repr preds e etuple attr } {
	global mr
	switch -- $kind {
		inputsuspension -
		outputsuspension -
		internal -
		verdict {
			set key [mr_find_or_mk_sublist $kind $chan]
			return [mr_do_do_insert $kind $chan $e $etuple $repr $key $attr]
		} input -
		output  {
			set key [mr_find_or_mk_visible_sublist $kind $chan $repr $preds]
			return [mr_do_do_insert $kind $chan $e $etuple $repr $key $attr]
		} default {
			debugMsg "mr_do_insert: internal error: unknown kind: $kind"
			return 0
		}
	}
}

proc mr_do_do_insert {kind chan e etuple repr key attr} {
	global mr
	set identical [lindex $etuple 6]
	debugMsg "mr_do_do_insert kind $kind chan $chan e $e id $identical ($etuple) \"$repr\" key $key"
	if {[string compare $identical ""] != 0} {
		if {[info exists mr(event,sublist,$kind,$chan,$identical,$key)]} {
			debugMsg "not inserting identical $kind $e: $identical (in menu sublist)"
			return 0
		} elseif {[info exists mr(event,$identical)]} {
			set i_k [lindex $mr(event,$identical) 0]
			set i_c [lindex $mr(event,$identical) 1]
			debugMsg "inserting identical $kind $e: $identical (elsewhere: $i_k $i_c)"
		} else {
			debugMsg "inserting identical $kind $e: $identical (unseen)"
		}
		set canonical $identical
	} else {
		if {[info exists mr(event,sublist,$kind,$chan,$e,$key)]} {
			debugMsg "not inserting same $kind $e (in menu sublist)"
			return 0
		}
		set canonical $e
	}
	lappend mr(sublist,$kind,$chan,$key) $e
	# 'register' the 'canonical version' of the event that we insert
	# (for subsequent identical searching)
	if {![info exists mr(event,sublist,$kind,$chan,$canonical,$key)]} {
		set mr(event,sublist,$kind,$chan,$canonical,$key) [list $kind $chan $key]
	}
	if {![info exists mr(event,$canonical)]} {
		set mr(event,$canonical) [list $kind $chan $key]
	}
	# register all details of the inserted event
	set mr(identical,$e) $identical
	set mr(repr,$kind,$chan,$e) $repr
	set mr(etuple,$kind,$chan,$e) $etuple
	set mr(attr,$kind,$chan,$e) $attr
	set free_vars [lindex $etuple 5]
	set fvl $free_vars
	set tmp {}
	while {[llength $fvl] > 0} {
		lappend tmp [list [lindex $fvl 0] [lindex $fvl 1] [lindex $fvl 2]]
		set fvl [lrange $fvl 3 end]
	}
	set mr(freevars,$kind,$chan,$e) $tmp
	return 1
}

proc mr_list_kind_chan { kind chan } {
	global mr
	set events {}
	foreach key $mr(keys,$kind,$chan) {
		switch -- $kind {
			inputsuspension -
			outputsuspension {
				set representative [lindex $mr(sublist,$kind,$chan,$key) 0]
				set repr $mr(repr,$kind,$chan,$representative)
			} input -
			  output {
				set representative [lindex $mr(sublist,$kind,$chan,$key) 0]
				set repr $mr(repr,$kind,$chan,$representative)
				# set free_vars $mr(freevars,$kind,$chan,$representative)
				# foreach fv $free_vars {
				###  regsub "wim1(\[^0-9\]|\$)"  $c "noot\\1" res
				#	regsub [lindex $fv 0] $repr "var_[lindex $fv 1]" repr
				#	regsub "[lindex $fv 0](\[^0-9\]|\$)" $repr "var_[lindex $fv 1]\\1" repr
				# }
			} default {
				# handle verdicts???
				set representative [lindex $mr(sublist,$kind,$chan,$key) 0]
				set repr $mr(repr,$kind,$chan,$representative)
			}
		}
		lappend events [list $kind $chan $repr $mr(sublist,$kind,$chan,$key) $mr(etuple,$kind,$chan,$representative) $mr(attr,$kind,$chan,$representative)]
	}
	return $events
}

proc mr_do_events {kinds} {
	global mr
	set events {}
	foreach kind $kinds {
		foreach chan $mr(channels,$kind) {
			foreach key $mr(keys,$kind,$chan) {
				lappend events $mr(sublist,$kind,$chan,$key)
			}
		}
	}
	return $events
}

# ==================================================

proc mr_find_or_mk_sublist {kind chan} {
	global mr
	set l [mr_sublists $kind $chan]
	if {[llength $l] > 0} {
		return [lindex $l 0]
	} else {
		set k [mr_mk_sublist $kind $chan]
		return $k
	}
}
proc mr_mk_sublist {kind chan} {
	global mr
	set k $mr(nextkey,$kind,$chan)
	lappend mr(keys,$kind,$chan) $k
	incr mr(nextkey,$kind,$chan)
	set mr(sublist,$kind,$chan,$k) {}
	return $k
}

proc mr_find_or_mk_visible_sublist {kind chan repr preds} {
	global mr
	if {[info exists mr(val2key,$kind,$chan,$repr,$preds)]} {
		debugMsg " mr_find_or_mk_visible_sublist found $kind $chan \"$repr\" $preds"
		return $mr(val2key,$kind,$chan,$repr,$preds)
	} else {
		debugMsg " mr_find_or_mk_visible_sublist new $kind $chan \"$repr\" $preds"
		set k [mr_mk_sublist $kind $chan]
		set mr(val2key,$kind,$chan,$repr,$preds) $k
		return $k
	}
}

proc mr_sublists {kind chan} {
	global mr
	return $mr(keys,$kind,$chan)
}
